.omemo-inactive-color { color: @insensitive_fg_color; }

.omemo-qr-not-available {color: red;}

.omemo-mono { font-size: 12px; font-family: monospace; }

.omemo-last-seen { font-size: 11px; }

.omemo-key-dialog scrolledwindow row {
    border-bottom: 1px solid;
    border-color: @unfocused_borders;
    padding: 10px 20px 10px 10px;
}
.omemo-key-dialog scrolledwindow row:last-child { border-bottom: 0px}
.omemo-key-dialog scrolledwindow { border: 1px solid; border-color:@unfocused_borders; }
.omemo-key-dialog list > row { outline: none; }

.omemo-trust-popover row { padding: 10px 15px 10px 10px; }
